// A polynomial formulae producing class with graph drawing facilities

        //////////////////////////////////////////////////////////
        // Written from scratch by me (Richard M Birch) aged 16 //
        //////////////////////////////////////////////////////////

// Version 1.0 July 1997

// Basically you give it some results e.g.
//
//    y = 1 when  x = 1
//    y = 8 when  x = 2
//    y = 27 when x = 3
//    y = 64 when x = 4
//
// in two double pointers - first one with the x values, second with the
// corresponding y values - and it will store the formula (x^3 or whatever
// it happens to be) in a structure that you can then access through various
// functions (or by returning the structure). You can then draw graphs of
// the function either on the screen or to a bitmap, and manipulate the axis
// however you want (well nearly).
// You access the formula with Numerator(int), Denominator(int) and
// Power(int). They return the coefficient (in fraction form incase it's
// more accurate) and the power of each x in the formula. You should call
// these functions Num_Results() amount of times to make sure you have the
// complete formula (although some coefficients may be zero - in the case of
// y = x^2, the coefficient of x^1 and x^0 will be zero).


// Disclaimer --- I accept NO responsibility for any effect this code has on
// yourself or anything to do with you!

// If you have any comments, suggestions or questions, e-mail me at
//
//       richardbirch@dial.pipex.com



//***************************************************************************

#include <allegro.h>  // Many thanks to Shawn Hargreaves

float int_power(float, int);
int cancel_fraction(double&, double&);

struct Formula {  // Used in Function
       int Num_Comps;            // no. of components in formula
       double X_Value[50];       // storage of original x values
       double Numerator[50];     // X-Coefficient's numerator
       double Denominator[50];   // X-Coeffienent's denominator
       int Power[50];            // Power of X and of coefficient
};

struct Func_Variables {    // used to store current state of Function - Call Function::State()
       int num_results, vis_step;
       int ERRORS;
       int grid_col, graph_col;
       double x_scale, y_scale;
       double min_x, max_x, min_y, max_y;
};

class Function {
      int num_results, vis_step, halved_x, halved_y;
      float alpha[51][50];
      Formula form;
      void get_formula(void);
      void init_form_variables(double*, double*);
      void init_func_variables(int);
      void check_zero(int, int);
      int ERRORS;                  // Non-zero if there has been an error
      int grid_col, graph_col;
      float x_scale, y_scale;
      float min_x, max_x, min_y, max_y;
public:
      Function(int, double*, double*); //Constructor that makes a formula
      Function(int);                   //Constructor that initiates the state variables
      Formula Form(void) { return form; } //Returns the formula being used in a struct
      void New_Formula(int, double*, double*); //Makes new formula
      void Copy_Formula(Formula);    //Allows you to give it it's own formula
      Func_Variables State(void);    //Returns the current state in a struct
      void Set_State(Func_Variables);  //Allows you to set your own state
      void Reset(void) { init_func_variables(num_results); }    //Resets to default state
      int Num_Results(void) { return(num_results); } //Returns the number of results given when the formula was made
      int Graph(void);      //draws the graph on the screen (presuming a gfx mode is set)
      int Graph(int, int, int, int);  //draws graph to screen given a set of axis x1, x2, y1, y2
      int Graph_Bmp(BITMAP*); //same as Graph but draws to a given bitmap
      int Graph_Bmp(BITMAP*, int, int, int, int); //guess!!
      void Set_Step(double step) { vis_step = int(step); if (!vis_step) vis_step++; } //sets step used in drawing the graph
      int Step(void) { return(vis_step); }  //returns the current step
      void Set_Colour(int grid, int graph) {  //sets the grid and curve colour
           grid_col = grid;
           graph_col = graph;
      }
      int Grid_Colour(void) { return(grid_col); }  //returns the current grid colour
      int Graph_Colour(void) { return(graph_col); } //guess!!
      void Set_Scale(double x, double y) {        //sets the scale drawn on the axis
           x_scale = x;
           y_scale = y;
      }
      double X_Scale(void) { return(x_scale); } //returns the current x-scale
      double Y_Scale(void) { return(y_scale); } //guess
      void Set_Min_Max(double, double, double, double);  //sets the axis min max - x1, x2, y1, y2
      void Check_Scale(void); //resets the scale to previous value if it's been changed automatically
      double Min_X(void) { return(min_x); }  //returns the min x-axis value
      double Max_X(void) { return(max_x); }
      double Min_Y(void) { return(min_y); }
      double Max_Y(void) { return(max_y); }
      int Error_Number(void) { return(ERRORS); } //returns zero if everythings OK
      double Numerator(int num) {  //returns 'form.Numerator[num]'
             if(num >= num_results)
                return(0);
             else return(form.Numerator[num]);
      }
      double Denominator(int num) {
             if(num >= num_results)
                return(0);
             else return(form.Denominator[num]);
      }
      float Result(float);  //returns y given x according to the formula
};

Function::Function(int num_res, double* x, double* res) {
     init_func_variables(num_res);
     init_form_variables(x, res);
     get_formula();
     form.X_Value;
}
Function::Function(int num_res) {
     num_results = num_res;
     init_func_variables(num_results);
//     init_form_variables(num_results);
}
void Function::New_Formula(int num_res, double* x, double* res) {
     num_results = num_res;
     init_form_variables(x, res);
     get_formula();
     form.X_Value;
}
void Function::Copy_Formula(Formula new_form) {
     form.Num_Comps = new_form.Num_Comps;
     num_results = form.Num_Comps;

     for (int count = 0; count != form.Num_Comps; count++) {
         form.Numerator[count] = new_form.Numerator[count];
         form.Denominator[count] = new_form.Denominator[count];
         form.Power[count] = new_form.Power[count];
     }
}
Func_Variables Function::State(void) {
     Func_Variables vars;
     vars.num_results = num_results;
     vars.vis_step = vis_step;
     vars.ERRORS = ERRORS;
     vars.grid_col = grid_col;
     vars.graph_col = graph_col;
     vars.x_scale = x_scale;
     vars.y_scale = y_scale;
     vars.min_x = min_x;
     vars.max_x = max_x;
     vars.min_y = min_y;
     vars.max_y = max_y;
     return(vars);
}
void Function::Set_State(Func_Variables vars) {
     num_results = vars.num_results;
     vis_step = vars.vis_step;
     ERRORS = vars.ERRORS;
     grid_col = vars.grid_col;
     graph_col = vars.graph_col;
     x_scale = vars.x_scale;
     y_scale = vars.y_scale;
     min_x = vars.min_x;
     max_x = vars.max_x;
     min_y = vars.min_y;
     max_y = vars.max_y;
}
void Function::init_func_variables(int num_res) {
     ERRORS = 0;
     grid_col = 4;            //normally red
     graph_col = 7;           //normally white
     halved_x = 0;
     halved_y = 0;
     num_results = num_res;
     vis_step = 2;
     x_scale = 2;
     y_scale = 2;
     min_x = -10; max_x = 10;
     min_y = -50; max_y = 50;
     form.Num_Comps = num_results;
}
void Function::init_form_variables(double* x, double* res) {
     form.Num_Comps = num_results;
     for (int count = 0; count != num_results; count++) {
         form.X_Value[count] = x[count];
         alpha[num_results][count] = res[count];
     }
}

void Function::get_formula(void) {  //this is the fundamental part of the class
// setup array of x, x^2... etc
     for(int count = 0; count != num_results; count++)
           for(int inter = 0; inter != num_results; inter++)
                   alpha[inter][count] =
                        int_power(form.X_Value[count], num_results - inter - 1);

// make useful bits in array
     for(int count = 0; count != num_results - 1; count++)
             for (int inter = num_results - 1; inter != count; inter--) {
                 check_zero (num_results - count - 1, inter);
                 for (int coeff = 0; coeff != num_results + 1; coeff++)
                     alpha[coeff][inter] -= alpha[coeff][inter - 1];
             }

// setup cof
     for(int count = 0; count != num_results; count++)
             form.Numerator[count] = alpha[num_results][count];

// get cof values
     for(int y = num_results - 1; y != -1; y--) {
             for(int take = 0; take != num_results - y - 1; take++)
                     form.Numerator[y] -= ((form.Numerator[num_results - 1 - take]
                                            / form.Denominator[num_results - 1 - take])
                                            * alpha[take][y]);

             form.Denominator[y] = alpha[num_results - y - 1][y];
             cancel_fraction(form.Numerator[y], form.Denominator[y]);
             form.Power[y] = y;
     }
}
// if the last two non-zero coefficients are not zero when subtacted then one
// must be multiplied by a suitable factor
void Function::check_zero(int x, int y) {
     if ((alpha[x][y] - alpha[x][y - 1])) {
        double mult_fac = (alpha[x][y - 1] / alpha[x][y]);
        for (int count = 0; count != num_results + 1; count++)
            alpha[count][y] *= mult_fac;
     }
}

float Function::Result(float x) {
       float result = 0;
       for (int pow = num_results; pow >= 1; pow--) {
           if (form.Denominator[pow - 1])
              result += ((form.Numerator[pow - 1] / form.Denominator[pow - 1])
                         * int_power(x, form.Power[pow - 1]));
       }
       return(result);
}
int Function::Graph(void) {
    double step = vis_step * ((max_x - min_x) / SCREEN_W);

    int scr_x = 0, scr_y = 0,
        last_scr_x = 0, last_scr_y = 0;

    int h_line = int(SCREEN_H * (max_y / (max_y - min_y)));
    int v_line = int(SCREEN_W - (SCREEN_W * (max_x / (max_x - min_x))));

    vline(screen, v_line, 0, SCREEN_H, grid_col);
    for(int y = 0; y <= SCREEN_H; y += int(SCREEN_H / ((max_y - min_y) / y_scale))) {
            if(y == SCREEN_H) y -= 1;
            hline(screen, v_line - 2, h_line + y, v_line + 2, grid_col);
            hline(screen, v_line - 2, h_line - y, v_line + 2, grid_col);
    }
    hline(screen, 0, h_line, SCREEN_W, grid_col);
    for(int x = 0; x <= SCREEN_W; x += int(SCREEN_W / ((max_x - min_x) / x_scale))) {
            if(x == SCREEN_W) x -= 1;
            vline(screen, v_line + x, h_line - 2, h_line + 2, grid_col);
            vline(screen, v_line - x, h_line - 2, h_line + 2, grid_col);
    }
    for(double real_x = min_x, real_y; scr_x < SCREEN_W; scr_x += vis_step, real_x += step) {
           real_y = Result(real_x);
           if (0 <= real_y < SCREEN_H) {
              scr_y = h_line - int((real_y / max_y) * (SCREEN_H * (max_y / (max_y - min_y))));
              if (real_x == min_x)
                 last_scr_y = scr_y;
              line(screen, scr_x, scr_y, scr_x - vis_step, last_scr_y, graph_col);
           }
           last_scr_x = scr_x;
           last_scr_y = scr_y;
    }
}
int Function::Graph(int x1, int x2, int y1, int y2) {
    Set_Min_Max(x1, x2, y1, y2);
    Graph();
}
int Function::Graph_Bmp(BITMAP* bmp) {
    clear(bmp);
    double step = vis_step * ((max_x - min_x) / bmp->w);

    int bmp_x = 0, bmp_y = 0,
        last_bmp_x = 0, last_bmp_y = 0;

    int h_line = int(bmp->h * (max_y / (max_y - min_y)));
    if(h_line == bmp->h) h_line -= 1;
    int v_line = int(bmp->w - (bmp->w * (max_x / (max_x - min_x))));
    if(v_line == bmp->w) v_line -= 1;

    vline(bmp, v_line, 0, bmp->h, grid_col);
    for(int y = 0; y <= bmp->h; y += (bmp->h / int((max_y - min_y) / y_scale))) {
            if(y == bmp->h) y -= 1;
            hline(bmp, v_line - 2, h_line + y, v_line + 2, grid_col);
            hline(bmp, v_line - 2, h_line - y, v_line + 2, grid_col);
    }
    hline(bmp, 0, h_line, bmp->w, grid_col);
    for(int x = 0; x <= bmp->w; x += (bmp->w / int((max_x - min_x) / x_scale))) {
            if(x == bmp->w) x -= 1;
            vline(bmp, v_line + x, h_line - 2, h_line + 2, grid_col);
            vline(bmp, v_line - x, h_line - 2, h_line + 2, grid_col);
    }
    for(double real_x = min_x, real_y; bmp_x < bmp->w; bmp_x += vis_step, real_x += step) {
           real_y = Result(real_x);
           if (0 <= real_y < SCREEN_H) {
              bmp_y = h_line - int((real_y / max_y) * (bmp->h * (max_y / (max_y - min_y))));
              if (real_x == min_x)
                 last_bmp_y = bmp_y;
              line(bmp, bmp_x, bmp_y, bmp_x - vis_step, last_bmp_y, graph_col);
           }
           last_bmp_x = bmp_x;
           last_bmp_y = bmp_y;
    }
}
int Function::Graph_Bmp(BITMAP* bmp, int x1, int x2, int y1, int y2) {
    Set_Min_Max(x1, x2, y1, y2);
    Graph_Bmp(bmp);
}
void Function::Set_Min_Max(double x1, double x2, double y1, double y2) {
     min_x = x1;
     if (((max_x = x2) <= x1))
        max_x = min_x + 1;
     if ((max_x - min_x) < x_scale) {
        x_scale /= 2;
        halved_x++;
     }
     min_y = y1;
     if (((max_y = y2) <= y1))
        max_y = min_y + 1;
     if ((max_y - min_y) < y_scale) {
        y_scale /= 2;
        halved_y++;
     }
}
void Function::Check_Scale(void) {
     if ((max_x - min_x) > (2 * x_scale))
        if (halved_x) {
           x_scale *= 2;
           halved_x--;
        }
     if ((max_y - min_y) > (2 * y_scale))
        if (halved_y) {
           y_scale *= 2;
           halved_y--;
        }
}

float int_power(float num, int exp) {
       float result = num;
       int neg = 0;
       if(exp == 0)
          return(1);
       if(exp < 0) {
          neg = 1;
          exp = 0 - exp;
       }
       for(int count = 1; count != exp; count++)
               result *= num;
       if(neg) return(1 / result);
       else return(result);
}

int primes[] = {
    2,
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
}; //first 10 prime numbers

int cancel_fraction(double& numerator, double& denominator) {
     double orig_num = numerator;
     if ((numerator / denominator) == int(numerator / denominator)) {
        numerator /= denominator;
        denominator = 1;
        return(1);
     }
     for (int count = 0; count != 10; count++)
         if (((numerator / primes[count]) == int(numerator / primes[count]))
             && ((denominator / primes[count]) == int(denominator / primes[count]))) {
                numerator /= primes[count];
                denominator /= primes[count];
         }
     if (orig_num == numerator)
        return(0);
     return(2);
}
